@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
  * USA
  */
 
@@ -30,12 +30,14 @@
 #include <gsf/gsf-timestamp.h>
 #include <gsf/gsf-meta-names.h>
 #include <gsf/gsf-doc-meta-data.h>
+#include <gsf/gsf-clip-data.h>
 
 #include <locale.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
 #include <time.h>
+#include <glib/gi18n-lib.h>
 
 #define NO_DEBUG_OLE_PROPS
 #ifndef NO_DEBUG_OLE_PROPS
@@ -273,6 +275,112 @@ msole_gsf_name_to_prop (char const *name)
 	return g_hash_table_lookup (name_to_prop_hash, (gpointer)name);
 }
 
+static void
+set_error_missing_data (GError **error, const char *property_name, gsize size_needed, gsize size_gotten)
+{
+	g_set_error (error,
+		     GSF_ERROR,
+		     GSF_ERROR_INVALID_DATA,
+		     _("Missing data when reading the %s property; got %" G_GSIZE_FORMAT "bytes, "
+		       "but %" G_GSIZE_FORMAT " bytes at least are needed."),
+		     property_name,
+		     size_needed,
+		     size_gotten);
+}
+
+/* Can return errors from gsf_blob_new() and GSF_ERROR_INVALID_DATA */
+static gboolean
+parse_vt_cf (GValue *res, guint8 const **data, guint8 const *data_end, GError **error)
+{
+	/* clipboard size		uint32		sizeof (clipboard format tag) + sizeof (clipboard data)
+	 * clipboard format tag		int32		see below
+	 * clipboard data		byte[]		see below
+	 *
+	 * Clipboard format tag:
+	 * -1 - Windows clipboard format
+	 * -2 - Macintosh clipboard format
+	 * -3 - GUID that contains a format identifier (FMTID)
+	 * >0 - custom clipboard format name plus data (see msdn site below)
+	 *  0 - No data
+	 *
+	 * References:
+	 * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/stg/stg/propvariant.asp
+	 * http://jakarta.apache.org/poi/hpsf/thumbnails.html
+	 * http://linux.com.hk/docs/poi/org/apache/poi/hpsf/Thumbnail.html
+	 * http://sparks.discreet.com/knowledgebase/public/solutions/ExtractThumbnailImg.htm
+	 */
+	guint32 clip_size, clip_data_size;
+	gint32 clip_format;
+	GsfBlob *blob;
+	GsfClipData *clip_data;
+
+	/* Clipboard size field */
+
+	if (data_end < *data + 4) {
+		set_error_missing_data (error, "VT_CF", 4, data_end - *data);
+		return FALSE;
+	}
+
+	clip_size = GSF_LE_GET_GUINT32 (*data);
+
+	if (clip_size < 4) {	/* must emcompass int32 format plus data size */
+		g_set_error (error,
+			     GSF_ERROR,
+			     GSF_ERROR_INVALID_DATA,
+			     _("Corrupt data in the VT_CF property; clipboard data length must be at least 4 bytes, "
+			       "but the data says it only has %" G_GSIZE_FORMAT " bytes available."),
+			     (gsize) clip_size);
+		return FALSE;
+	}
+
+	*data += 4;
+
+	/* Check clipboard format plus data size */
+
+	if (data_end < *data + clip_size) {
+		set_error_missing_data (error, "VT_CF", clip_size, data_end - *data);
+		return FALSE;
+	}
+
+	clip_format = GSF_LE_GET_GINT32 (*data);
+	*data += 4;
+
+	switch (clip_format) {
+	case GSF_CLIP_FORMAT_WINDOWS_CLIPBOARD:
+	case GSF_CLIP_FORMAT_MACINTOSH_CLIPBOARD:
+	case GSF_CLIP_FORMAT_GUID:
+	case GSF_CLIP_FORMAT_NO_DATA:
+		/* everything is ok */
+		break;
+
+	default:
+		if (clip_format > 0)
+			clip_format = GSF_CLIP_FORMAT_CLIPBOARD_FORMAT_NAME;
+		else
+			clip_format = GSF_CLIP_FORMAT_UNKNOWN;
+
+		break;
+	}
+
+	clip_data_size = clip_size - 4;
+
+	blob = gsf_blob_new (clip_data_size, *data, error);
+
+	*data += clip_data_size;
+
+	if (!blob)
+		return FALSE;
+
+	clip_data = gsf_clip_data_new (clip_format, blob);
+	g_object_unref (blob);
+
+	g_value_init (res, GSF_TYPE_CLIP_DATA);
+	g_value_set_object (res, clip_data);
+	g_object_unref (clip_data);
+
+	return TRUE;
+}
+
 static GValue *
 msole_prop_parse (GsfMSOleMetaDataSection *section,
 		  guint32 type, guint8 const **data, guint8 const *data_end)
@@ -280,8 +388,8 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 	GValue *res;
 	char *str;
 	guint32 len;
-	gsize gslen;
 	gboolean const is_vector = type & VT_VECTOR;
+	GError *error;
 
 	g_return_val_if_fail (!(type & (unsigned)(~0x1fff)), NULL); /* not valid in a prop set */
 
@@ -388,7 +496,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * 64-bit floating-point number representing the number of days
 		 * (not seconds) since December 31, 1899.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		break;
 
 	case VT_BSTR :
@@ -397,11 +505,11 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * ceeded by a DWORD representing the byte count of the number
 		 * of bytes in the string (including the  terminating null).
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		break;
 
 	case VT_DISPATCH :
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		break;
 
 	case VT_BOOL :
@@ -451,7 +559,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		/* 8-byte signed integer */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, G_TYPE_INT64);
-		g_value_set_uint (res, GSF_LE_GET_GINT64 (*data));
+		g_value_set_int64 (res, GSF_LE_GET_GINT64 (*data));
 		*data += 8;
 		break;
 
@@ -459,7 +567,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		/* 8-byte unsigned integer */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, G_TYPE_UINT64);
-		g_value_set_uint (res, GSF_LE_GET_GUINT64 (*data));
+		g_value_set_uint64 (res, GSF_LE_GET_GUINT64 (*data));
 		*data += 8;
 		break;
 
@@ -478,16 +586,23 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		g_return_val_if_fail (len < 0x10000, NULL);
 		g_return_val_if_fail (*data + 4 + len*section->char_size <= data_end, NULL);
 
-		gslen = 0;
+		error = NULL;
+		d (gsf_mem_dump (*data + 4, len * section->char_size););
 		str = g_convert_with_iconv (*data + 4,
 			len * section->char_size,
-			section->iconv_handle, &gslen, NULL, NULL);
-		len = (guint32)gslen;
+			section->iconv_handle, NULL, NULL, &error);
 
 		g_value_init (res, G_TYPE_STRING);
-		g_value_set_string (res, str);
-		g_free (str);
-		*data += 4 + len;
+		if (NULL != str) {
+			g_value_set_string (res, str);
+			g_free (str);
+		} else if (NULL != error) {
+			g_warning ("error: %s", error->message);
+			g_error_free (error);
+		} else {
+			g_warning ("unknown error converting string property, using blank");
+		}
+		*data += 4 + len * section->char_size;
 		break;
 
 	case VT_LPWSTR :
@@ -505,14 +620,22 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		g_return_val_if_fail (len < 0x10000, NULL);
 		g_return_val_if_fail (*data + 4 + len <= data_end, NULL);
 
+		error = NULL;
+		d (gsf_mem_dump (*data + 4, len*2););
 		str = g_convert (*data + 4, len*2,
-				 "UTF-8", "UTF-16LE", &gslen, NULL, NULL);
-		len = (guint32)gslen;
+				 "UTF-8", "UTF-16LE", NULL, NULL, &error);
 
 		g_value_init (res, G_TYPE_STRING);
-		g_value_set_string (res, str);
-		g_free (str);
-		*data += 4 + len;
+		if (NULL != str) {
+			g_value_set_string (res, str);
+			g_free (str);
+		} else if (NULL != error) {
+			g_warning ("error: %s", error->message);
+			g_error_free (error);
+		} else {
+			g_warning ("unknown error converting string property, using blank");
+		}
+		*data += 4 + len*2;
 		break;
 
 	case VT_FILETIME :
@@ -524,11 +647,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		GsfTimestamp ts;
 
 		ft /= 10000000; /* convert to seconds */
-#ifdef _MSC_VER
-		ft -= 11644473600i64; /* move to Jan 1 1970 */
-#else
-		ft -= 11644473600ULL; /* move to Jan 1 1970 */
-#endif
+		ft -= G_GINT64_CONSTANT (11644473600); /* move to Jan 1 1970 */
 		ts.timet = (time_t)ft;
 		g_value_init (res, GSF_TIMESTAMP_TYPE);
 		gsf_value_set_timestamp (res, &ts);
@@ -544,7 +663,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * tation of a VT_BLOB is similar to that of a VT_BSTR but does
 		 * not guarantee a null byte at the end of the data.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		g_free (res);
 		res = NULL;
 		break;
@@ -556,7 +675,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * in the format of a serialized VT_LPSTR, which names the stream
 		 * containing the data.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		g_free (res);
 		res = NULL;
 		break;
@@ -568,7 +687,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * in the format of a serialized VT_LPSTR, which names the
 		 * IStorage containing the data.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		g_free (res);
 		res = NULL;
 		break;
@@ -579,7 +698,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * serialized object, which is a class ID followed by initiali-
 		 * zation data for the class.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		g_free (res);
 		res = NULL;
 		break;
@@ -589,7 +708,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * Same as VT_STORAGE, but indicates that the designated IStorage
 		 * contains a loadable object.
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
 		g_free (res);
 		res = NULL;
 		break;
@@ -603,23 +722,21 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		 * which is in the format of a class ID followed by initialization
 		 * data for that class
 		 */
-#warning TODO
+/* FIXME FIXME FIXME  TODO */
+
 		g_free (res);
 		res = NULL;
 		break;
 
 	case VT_CF :
-		/*
-		 * a BLOB containing a clipboard format identifier followed by
-		 * the data in that format.  That is, following the VT_CF tag is
-		 * data in the format of a VT_BLOB: a CWORD count of bytes,
-		 * followed by that many bytes of data in the format of a packed
-		 * VTCFREP, followed immediately by an array of bytes as appropriate
-		 * for data in the clipboard format.
-		 */
-#warning TODO
-		g_free (res);
-		res = NULL;
+		error = NULL;
+		if (!parse_vt_cf (res, data, data_end, &error)) {
+			/* suck, we can't propagate the error upwards */
+			g_warning ("error: %s", error->message);
+			g_error_free (error);
+			g_free (res);
+			res = NULL;
+		}
 		break;
 
 	case VT_CLSID :
@@ -662,12 +779,14 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 			g_free (val);
 		});
 	} else {
-		char const *type_name = msole_vt_name (type);
-		if (type_name) {
-			g_print ("A '%s' property could not be parsed\n", type_name);
-		} else {
-			g_print ("A %d property could not be parsed\n", type);
-		}
+		d ({
+			char const *type_name = msole_vt_name (type);
+			if (type_name) {
+				g_print ("A '%s' property could not be parsed\n", type_name);
+			} else {
+				g_print ("A %d property could not be parsed\n", type);
+			}
+		});
 	}
 	return res;
 }
@@ -782,9 +901,15 @@ msole_prop_read (GsfInput *in,
 static int
 msole_prop_cmp (gconstpointer a, gconstpointer b)
 {
-	GsfMSOleMetaDataProp const *prop_a = a ;
-	GsfMSOleMetaDataProp const *prop_b = b ;
-	return prop_a->offset - prop_b->offset;
+	GsfMSOleMetaDataProp const *prop_a = a;
+	GsfMSOleMetaDataProp const *prop_b = b;
+
+	if (prop_a->offset < prop_b->offset)
+		return -1;
+	else if (prop_a->offset > prop_b->offset)
+		return +1;
+	else
+		return 0;
 }
 
 /**
@@ -1059,8 +1184,8 @@ msole_metadata_write_prop (WritePropState *state,
 		gsf_output_write (state->out, 4, buf);
 	}
 	if (NULL != map && map->prefered_type != type) {
-		g_print ("Exporting property '%s' with type 0x%x rather than the usual 0x%x\n",
-			 map->gsf_name, type, map->prefered_type);
+		d(g_print ("Exporting property '%s' with type 0x%x rather than the usual 0x%x\n",
+			   map->gsf_name, type, map->prefered_type););
 	}
 
 	if (type & VT_VECTOR) {
@@ -1106,9 +1231,9 @@ msole_metadata_write_prop (WritePropState *state,
 		return gsf_output_write (state->out, 8, buf);
 
 	case VT_LPSTR : {
-#warning TODO : use iconv from codepage
+/* FIXME FIXME FIXME  TODO : use iconv from codepage */
 		char const *txt = g_value_get_string (value);
-		unsigned len = strlen (txt);
+		unsigned len = (NULL != txt) ? strlen (txt) : 0;
 		GSF_LE_SET_GUINT32 (buf, len+1);
 		return  gsf_output_write (state->out, 4, buf) &&
 			gsf_output_write (state->out, len, txt) &&
@@ -1117,18 +1242,15 @@ msole_metadata_write_prop (WritePropState *state,
 
 	case VT_FILETIME : {
 		GsfTimestamp const *ts = g_value_get_boxed (value);
-		guint64 timet_value = ts->timet;
+		gint32  timet_signed = (gint32) ts->timet;
+		guint64 ft;
 
-#ifdef _MSC_VER
-		timet_value += 11644473600i64;
-#else
-		timet_value += 11644473600ULL;
-#endif
-		timet_value *= 10000000;
+		ft = timet_signed + G_GINT64_CONSTANT (11644473600);
+		ft *= 10000000;
 
-		GSF_LE_SET_GUINT64 (buf+4, timet_value);
+		GSF_LE_SET_GUINT64 (buf, ft);
 
-		return gsf_output_write (state->out, 12, buf);
+		return gsf_output_write (state->out, 8, buf);
 	}
 
 	default:
@@ -1900,6 +2022,7 @@ gsf_msole_iconv_get_codepage_string_list (int codepage)
 		case 1201:
 			cp_list = g_slist_prepend (cp_list, g_strdup ("UTF-16BE"));
 			break;
+		case 0x8000:
 		case 10000:
 			cp_list = g_slist_prepend (cp_list, g_strdup ("MACROMAN"));
 			cp_list = g_slist_prepend (cp_list, g_strdup ("MACINTOSH"));
@@ -1908,6 +2031,9 @@ gsf_msole_iconv_get_codepage_string_list (int codepage)
 		case 65001:
 			cp_list = g_slist_prepend (cp_list, g_strdup ("UTF-8"));
 			break;
+		case 0x8001:
+			/* according to OOo docs 8001 is a synonym CP1252 */
+			codepage = 1252;
 		default:
 			cp_list = g_slist_prepend (cp_list, g_strdup_printf ("CP%u", codepage));
 	}
